import logging
from typing import Type
from unittest.mock import patch

from azure.kusto.data import KustoClient, KustoConnectionStringBuilder
from azure.kusto.data.exceptions import KustoError

from pykusto import PyKustoClient, column_generator as col, Query, KustoServiceError, RetryConfig, NO_RETRIES, ClientRequestProperties
# noinspection PyProtectedMember
from pykusto._src.logger import _logger
# noinspection PyProtectedMember
from test.test_base import TestBase, MockKustoClient, RecordedQuery, mock_response


class TestClient(TestBase):
    def test_single_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['mock_table']
        Query(table).take(5).execute()
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')],
            mock_kusto_client.recorded_queries
        )

    def test_execute_no_table(self):
        self.assertRaises(
            RuntimeError("No table supplied"),
            Query().take(5).execute
        )

    def test_execute_already_bound(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['mock_table']

        self.assertRaises(
            RuntimeError("This table is already bound to a query"),
            Query(table).take(5).execute,
            table
        )

    def test_single_table_on_execute(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['mock_table']
        Query().take(5).execute(table)
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')],
            mock_kusto_client.recorded_queries,
        )

    def test_get_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db'].get_table('mock_table')
        Query(table).take(5).execute()
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')],
            mock_kusto_client.recorded_queries,
        )

    def test_union_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db'].get_table('test_table1', 'test_table2')
        Query(table).take(5).execute()
        self.assertEqual(
            [RecordedQuery('test_db', 'union test_table1, test_table2 | take 5')],
            mock_kusto_client.recorded_queries,
        )

    def test_union_table_with_wildcard(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table_*']
        Query(table).take(5).execute()
        self.assertEqual(
            [RecordedQuery('test_db', 'union test_table_* | take 5')],
            mock_kusto_client.recorded_queries,
        )

    def test_default_authentication(self):
        mock_kusto_client = MockKustoClient()
        with patch('pykusto._src.client.PyKustoClient._get_client_for_cluster', lambda s: mock_kusto_client):
            table = PyKustoClient('https://help.kusto.windows.net/')['test_db']['mock_table']
            Query().take(5).execute(table)
        self.assertIs(
            mock_kusto_client,
            table._Table__database._Database__client._PyKustoClient__client,
        )
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')],
            mock_kusto_client.recorded_queries,
        )

    def test_client_instances(self):
        with patch('pykusto._src.client.PyKustoClient._get_client_for_cluster', MockKustoClient):
            client_1 = PyKustoClient('https://help.kusto.windows.net/')
            client_2 = PyKustoClient('https://help.kusto.windows.net/')

        self.assertIsNot(
            client_1._PyKustoClient__client,
            client_2._PyKustoClient__client,
        )

    def test_client_instances_cached(self):
        with patch('pykusto._src.client.PyKustoClient._get_client_for_cluster', MockKustoClient):
            client_1 = PyKustoClient('https://help.kusto.windows.net/', use_global_cache=True)
            client_2 = PyKustoClient('https://help.kusto.windows.net/', use_global_cache=True)

        self.assertIs(
            client_1._PyKustoClient__client,
            client_2._PyKustoClient__client,
        )

    def test_client_instances_cached_distinct(self):
        with patch('pykusto._src.client.PyKustoClient._get_client_for_cluster', MockKustoClient):
            client_1 = PyKustoClient('https://help1.kusto.windows.net/', use_global_cache=True)
            client_2 = PyKustoClient('https://help2.kusto.windows.net/', use_global_cache=True)

        self.assertIsNot(
            client_1._PyKustoClient__client,
            client_2._PyKustoClient__client,
        )

    def test_cross_cluster_join(self):
        client1 = MockKustoClient("https://one.kusto.windows.net")
        client2 = MockKustoClient("https://two.kusto.windows.net")
        table1 = PyKustoClient(client1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(client2)['test_db_2']['test_table_2']
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [RecordedQuery('test_db_1', 'test_table_1 | take 5 | join  (cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2") | take 6) on foo')],
            client1.recorded_queries,
        )

    def test_cross_cluster_join_with_union(self):
        client1 = MockKustoClient("https://one.kusto.windows.net")
        client2 = MockKustoClient("https://two.kusto.windows.net")
        table1 = PyKustoClient(client1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(client2)['test_db_2'].get_table('test_table_2_*')
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [RecordedQuery('test_db_1', 'test_table_1 | take 5 | join  (union cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2_*") | take 6) on foo')],
            client1.recorded_queries,
        )

    def test_cross_cluster_join_with_union_2(self):
        client1 = MockKustoClient("https://one.kusto.windows.net")
        client2 = MockKustoClient("https://two.kusto.windows.net")
        table1 = PyKustoClient(client1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(client2)['test_db_2'].get_table('test_table_2_*', 'test_table_3_*')
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [RecordedQuery(
                'test_db_1',
                'test_table_1 | take 5 | join  (union cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2_*"), '
                'cluster("two.kusto.windows.net").database("test_db_2").table("test_table_3_*") | take 6) on foo',
            )],
            client1.recorded_queries,
        )

    def test_client_for_cluster_with_azure_cli_auth(self):
        with self.assertLogs(_logger, logging.INFO) as cm:
            client = PyKustoClient('https://help.kusto.windows.net', fetch_by_default=False)
            self.assertIsInstance(client._PyKustoClient__client, KustoClient)
            self.assertEqual('https://help.kusto.windows.net', client.get_cluster_name())
        self.assertEqual([], cm.output)

    def test_client_for_cluster_with_aad_device_auth(self):
        with self.assertLogs(_logger, logging.INFO) as cm:
            client = PyKustoClient('https://help.kusto.windows.net', fetch_by_default=False, auth_method=KustoConnectionStringBuilder.with_aad_device_authentication)
            self.assertIsInstance(client._PyKustoClient__client, KustoClient)
            self.assertEqual('https://help.kusto.windows.net', client.get_cluster_name())
        self.assertEqual([], cm.output)

    def test_request_properties(self):
        properties = ClientRequestProperties()
        properties.set_option(ClientRequestProperties.results_defer_partial_query_failures_option_name, False)
        properties.set_parameter('xIntValue', 11)

        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['mock_table']
        Query(table).take(5).execute(properties=properties)
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5', properties)],
            mock_kusto_client.recorded_queries
        )

    @staticmethod
    def unreliable_mock_kusto_client(number_of_failures: int, exception_type: Type[Exception] = KustoServiceError):
        TestClient.attempt = 1

        def main_response():
            if TestClient.attempt <= number_of_failures:
                TestClient.attempt += 1
                raise exception_type("Mock exception for test", None)
            return mock_response(tuple())()

        return MockKustoClient(main_response=main_response)

    def retries_base(self, number_of_retries: int):
        mock_kusto_client = self.unreliable_mock_kusto_client(number_of_retries - 1)
        table = PyKustoClient(mock_kusto_client, fetch_by_default=False, retry_config=RetryConfig(number_of_retries, sleep_time=0.1, jitter=0))['test_db']['mock_table']
        with self.assertLogs(_logger, logging.INFO) as cm:
            Query(table).take(5).execute()
        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')] * number_of_retries,
            mock_kusto_client.recorded_queries
        )
        self.assertEqual(
            [
                f"INFO:pykusto:Attempt number {i} out of {number_of_retries} failed, "
                f"previous sleep time was 0.1 seconds. Exception: KustoServiceError('Mock exception for test')"
                for i in range(1, number_of_retries)
            ],
            cm.output
        )

    def test_2_retry(self):
        self.retries_base(2)

    def test_3_retries(self):
        self.retries_base(3)

    def test_missing_retries(self):
        mock_kusto_client = self.unreliable_mock_kusto_client(1)
        table = PyKustoClient(
            mock_kusto_client, fetch_by_default=False, retry_config=NO_RETRIES
        )['test_db']['mock_table']
        self.assertRaises(
            KustoServiceError("Mock exception for test", None),
            lambda: Query(table).take(5).execute(),
        )

    def test_non_transient_exception(self):
        mock_kusto_client = self.unreliable_mock_kusto_client(1, KustoError)
        table = PyKustoClient(mock_kusto_client, fetch_by_default=False, retry_config=RetryConfig(2, sleep_time=0.1, jitter=0))['test_db']['mock_table']
        self.assertRaises(
            KustoError("Mock exception for test", None),
            lambda: Query(table).take(5).execute(),
        )

    def test_override_retry_config_for_query(self):
        num_retries = 2
        num_failed_attempts = num_retries - 1
        retry_sleep_time = 0.1

        mock_kusto_client = self.unreliable_mock_kusto_client(num_failed_attempts)
        table = PyKustoClient(mock_kusto_client, fetch_by_default=False)['test_db']['mock_table']
        retry_config = RetryConfig(num_retries, sleep_time=retry_sleep_time, jitter=0)

        with self.assertLogs(_logger, logging.INFO) as cm:
            Query(table).take(5).execute(retry_config=retry_config)

        self.assertEqual(
            [RecordedQuery('test_db', 'mock_table | take 5')] * num_retries,
            mock_kusto_client.recorded_queries
        )
        self.assertEqual(
            [
                f"INFO:pykusto:Attempt number {num_failed_attempts} out of {num_retries} failed, "
                f"previous sleep time was {retry_sleep_time} seconds. "
                f"Exception: KustoServiceError('Mock exception for test')"
            ],
            cm.output
        )



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.install import install

assert sys.version_info[0] == 3
__version__ = None
with open(os.path.join('.', 'pykusto', '__init__.py')) as f:
    for line in f:
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            __version__ = line.split(delim)[1]
assert __version__ is not None, 'Unable to determine version'

core_requires = [
    'redo==2.0.4',
]

non_pyspark_requires = [
    # Not required in PySpark, because authentication is handled differently there.
    # Release notes: https://github.com/Azure/azure-kusto-python/releases
    'azure-kusto-data==2.1.1',  # Earlier versions not supported because of: https://github.com/Azure/azure-kusto-python/issues/312
]

# Not required in PySpark because it is already installed there.
# pandas release notes: https://pandas.pydata.org/docs/whatsnew/index.html
# Tests use DataFrame constructor options introduced in 0.25.0
if sys.version_info[1] <= 6:
    # pandas support for Python 3.6 was dropped starting from version 1.2.0
    non_pyspark_requires.append('pandas>=0.25.0,<1.2.0')
    # In numpy the support was dropped in 1.20.0, and also the transitive dependency in pandas is not correctly restricted
    non_pyspark_requires.append('numpy<1.20.0')
else:
    non_pyspark_requires.append('pandas>=0.25.0,<=1.2.4')


# Allows installing with '--pyspark' to avoid unneeded dependencies.
# Usage:
#   pip install pykusto --global-option pyspark
# OR
#   python setup.py install --pyspark
class CustomInstall(install):
    user_options = install.user_options + [('pyspark', None, None)]

    def initialize_options(self):
        super().initialize_options()
        # noinspection PyAttributeOutsideInit
        self.pyspark = None

    def run(self):
        if self.pyspark:
            # noinspection PyUnresolvedReferences
            self.distribution.install_requires = core_requires
        super().run()


setup(
    name='pykusto',
    version=__version__,
    packages=find_packages(exclude=['test']),
    url='https://github.com/Azure/pykusto',
    license='MIT License',
    author='Microsoft Corporation',
    author_email='yomost@microsoft.com',
    description='Advanced python SDK for Azure Data Explorer',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    keywords="kusto azure-data-explorer client library query",
    cmdclass={'install': CustomInstall},
    install_requires=core_requires + non_pyspark_requires,
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'flake8',
            'typeguard',
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        # Make sure this list matches the tested versions in runtests.yml
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
)



from multiprocessing import Lock
from typing import Iterable, Callable, Dict, Union, Optional
from urllib.parse import urlparse

import pandas as pd
from azure.kusto.data import KustoClient, KustoConnectionStringBuilder
from azure.kusto.data.exceptions import KustoServiceError
from azure.kusto.data.helpers import dataframe_from_result_table
from azure.kusto.data.response import KustoResponseDataSet

from .client_base import KustoResponseBase, PyKustoClientBase, RetryConfig, NO_RETRIES, ClientRequestProperties
from .kql_converters import KQL


class KustoResponse(KustoResponseBase):
    __response: KustoResponseDataSet

    def __init__(self, response: KustoResponseDataSet):
        self.__response = response

    def get_rows(self) -> Iterable[Iterable]:
        return self.__response.primary_results[0].rows

    def to_dataframe(self) -> pd.DataFrame:
        return dataframe_from_result_table(self.__response.primary_results[0])


class PyKustoClient(PyKustoClientBase):
    """
    Handle to a Kusto cluster.
    Uses :class:`ItemFetcher` to fetch and cache the full cluster schema, including all databases, tables, columns and
    their types.
    """
    __client: KustoClient
    __auth_method: Callable[[str], KustoConnectionStringBuilder]

    # Static members
    __global_client_cache: Dict[str, KustoClient] = {}
    __global_cache_lock: Lock = Lock()

    def __init__(
            self, client_or_cluster: Union[str, KustoClient], fetch_by_default: bool = True, use_global_cache: bool = False,
            retry_config: RetryConfig = NO_RETRIES,
            auth_method: Optional[Callable[[str], KustoConnectionStringBuilder]] = KustoConnectionStringBuilder.with_az_cli_authentication,
    ) -> None:
        """
        Create a new handle to a Kusto cluster. The value of "fetch_by_default" is used for current instance, and also passed on to database instances.

        :param client_or_cluster: Either a KustoClient instance, or a cluster URL. In case a cluster URL is provided, a KustoClient is generated using the provided auth_method.
        :param use_global_cache: If true, share a global client cache between all instances. Provided for convenience during development, not recommended for general use.
        :param retry_config: An instance of RetryConfig which instructs the client how to perform retries in case of failure. The default is NO_RETRIES.
        :param auth_method: A method that returns a KustoConnectionStringBuilder for authentication. The default is 'KustoConnectionStringBuilder.with_az_cli_authentication'.
        A popular alternative is 'KustoConnectionStringBuilder.with_aad_device_authentication'
        """
        self.__auth_method = auth_method
        client_resolved = False
        if isinstance(client_or_cluster, KustoClient):
            self.__client = client_or_cluster
            client_resolved = True
            # noinspection PyProtectedMember
            cluster_name = urlparse(client_or_cluster._query_endpoint).netloc
            assert not use_global_cache, "Global cache not supported when providing your own client instance"
        else:
            cluster_name = client_or_cluster
        super().__init__(cluster_name, fetch_by_default, retry_config)
        if not client_resolved:
            self.__client = (self._cached_get_client_for_cluster if use_global_cache else self._get_client_for_cluster)()

    def __repr__(self) -> str:
        return f"PyKustoClient('{self._cluster_name}')"

    def _internal_execute(self, database: str, query: KQL, properties: ClientRequestProperties = None, retry_config: RetryConfig = None) -> KustoResponse:
        resolved_retry_config = self._retry_config if retry_config is None else retry_config
        if resolved_retry_config is not None:
            resolved_retry_config = resolved_retry_config.retry_on(KustoServiceError)
        return KustoResponse(resolved_retry_config.retry(lambda: self.__client.execute(database, query, properties)))

    def _get_client_for_cluster(self) -> KustoClient:
        return KustoClient(self.__auth_method(self._cluster_name))

    def _cached_get_client_for_cluster(self) -> KustoClient:
        """
        Provided for convenience during development, not recommended for general use.
        """
        with PyKustoClient.__global_cache_lock:
            client = PyKustoClient.__global_client_cache.get(self._cluster_name)
            if client is None:
                client = self._get_client_for_cluster()
                PyKustoClient.__global_client_cache[self._cluster_name] = client
                assert len(PyKustoClient.__global_client_cache) <= 1024, "Global client cache cannot exceed size of 1024"

        return client



from datetime import datetime, timedelta
from enum import Enum
from typing import Union, Mapping, Type, Dict, Callable, Tuple, List, Set, FrozenSet, Any

PythonTypes = Union[str, int, float, bool, datetime, Mapping, List, Tuple, timedelta]


class _KustoType(Enum):
    """
    https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/
    Order of entries determines preference when converting Python to Kusto types
    """
    BOOL = ('bool', 'I8', 'System.SByte', bool)
    DATETIME = ('datetime', 'DateTime', 'System.DateTime', datetime)
    ARRAY = ('dynamic', 'Dynamic', 'System.Object', List, Tuple)
    MAPPING = ('dynamic', 'Dynamic', 'System.Object', Mapping)
    LONG = ('long', 'I64', 'System.Int64', int)
    INT = ('int', 'I32', 'System.Int32', int)
    REAL = ('real', 'R64', 'System.Double', float)
    STRING = ('string', 'StringBuffer', 'System.String', str)
    TIMESPAN = ('timespan', 'TimeSpan', 'System.TimeSpan', timedelta)
    DECIMAL = ('decimal', 'Decimal', 'System.Data.SqlTypes.SqlDecimal', int)
    GUID = ('guid', 'UniqueId', 'System.Guid')  # Not supported by Kusto yet

    # Deprecated types, kept here for back compatibility
    # https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/scalar-data-types/unsupported-data-types
    FLOAT = ('float', 'R32', 'System.Single', float)
    INT16 = ('int16', 'I16', 'System.Int16', int)
    UINT16 = ('uint16', 'UI16', 'System.UInt16', int)
    UINT32 = ('uint32', 'UI32', 'System.UInt32', int)
    UINT64 = ('uint64', 'UI64', 'System.UInt64', int)
    UINT8 = ('uint8', 'UI8', 'System.Byte', int)

    primary_name: str
    internal_name: str
    dot_net_name: str
    python_types: Tuple[PythonTypes, ...]

    def __init__(self, primary_name: str, internal_name: str, dot_net_name: str, *python_types: PythonTypes) -> None:
        self.primary_name = primary_name
        self.internal_name = internal_name
        self.dot_net_name = dot_net_name
        self.python_types = python_types

    def is_type_of(self, obj: Any) -> bool:
        for python_type in self.python_types:
            if python_type is not None and isinstance(obj, python_type):
                return True
        return False

    def is_superclass_of(self, t: Type) -> bool:
        for python_type in self.python_types:
            if python_type is not None and issubclass(t, python_type):
                return True
        return False


_INTERNAL_NAME_TO_TYPE: Dict[str, _KustoType] = {t.internal_name: t for t in _KustoType}
_DOT_NAME_TO_TYPE: Dict[str, _KustoType] = {t.dot_net_name: t for t in _KustoType}
_NUMBER_TYPES: FrozenSet[_KustoType] = frozenset([
    _KustoType.INT, _KustoType.LONG, _KustoType.REAL, _KustoType.DECIMAL, _KustoType.FLOAT,
    _KustoType.INT16, _KustoType.UINT16, _KustoType.UINT32, _KustoType.UINT64, _KustoType.UINT8
])
_PYTHON_TYPE_TO_KUSTO_TYPE: Dict[Type, _KustoType] = {
    python_type: kusto_type
    for kusto_type in reversed(_KustoType)
    for python_type in kusto_type.python_types
}


class _TypeRegistrar:
    """
    A factory for annotations that are used to create a mapping between Kusto types and python types / functions.
    Each annotation must be called with a Kusto type as a parameter. The `for_obj` and `for_type` methods
    can then be used to retrieve the python type or function corresponding to a given Kusto type.
    """
    name: str
    registry: Dict[_KustoType, Union[Type, Callable]]

    def __init__(self, name: str) -> None:
        """
        :param name: Name is used for better logging and clearer errors
        """
        self.name = name
        self.registry = {}

    def __repr__(self) -> str:
        return self.name

    def __call__(self, *types: _KustoType) -> Callable[[Union[Type, Callable]], Union[Type, Callable]]:
        def inner(wrapped: Union[Type, Callable]) -> Union[Type, Callable]:
            for t in types:
                previous = self.registry.setdefault(t, wrapped)
                if previous is not wrapped:
                    raise TypeError(f"{self}: type already registered: {t.primary_name}")
            return wrapped

        return inner

    def for_obj(self, obj: PythonTypes) -> Any:
        """
        Given an object of Kusto type, retrieve the python type or function associated with the object's type, and call
        it with the given object as a parameter

        :param obj: An object of Kusto type
        :return: Associated python object
        """
        for registered_type, registered_callable in self.registry.items():
            if registered_type.is_type_of(obj):
                return registered_callable(obj)
        raise ValueError(f"{self}: no registered callable for object {obj} of type {type(obj).__name__}")

    def for_type(self, t: Type) -> Union[Type, Callable]:
        """
        Given a Kusto type, retrieve the associated python type or function

        :param t: A Kusto type
        :return: Associated python object
        """
        for registered_type, registered_callable in self.registry.items():
            if registered_type.is_superclass_of(t):
                return registered_callable
        raise ValueError(f"{self}: no registered callable for type {t.__name__}")

    def inverse(self, target_callable: Any) -> Set[_KustoType]:
        result: Set[_KustoType] = set()
        for kusto_type, associated_callable in self.registry.items():
            if isinstance(target_callable, associated_callable):
                result.add(kusto_type)
        return result

    def get_base_types(self, obj: Any) -> Set[_KustoType]:
        """
        For a given object, return the associated basic type, which is a member of :class:`KustoType`

        :param obj: The given object for which the type is resolved
        :return: A type which is a member of `KustoType`
        """
        for kusto_type in _KustoType:
            if kusto_type.is_type_of(obj):
                # The object is already a member of Kusto types
                return {kusto_type}
        # The object is one of the expression types decorated with a TypeRegistrar, therefore the original types are
        base_types: Set[_KustoType] = self.inverse(obj)
        assert len(base_types) > 0, f"get_base_types called for unsupported type: {type(obj).__name__}"
        return base_types

    def assert_all_types_covered(self) -> None:
        missing = set(t for t in _KustoType if len(t.python_types) > 0) - set(self.registry.keys())
        assert len(missing) == 0, [t.name for t in missing]


_kql_converter = _TypeRegistrar("KQL Converter")
_typed_column = _TypeRegistrar("Column")
_plain_expression = _TypeRegistrar("Plain expression")
_aggregation_expression = _TypeRegistrar("Aggregation expression")


def _get_base_types(obj: Any) -> Set[_KustoType]:
    """
    A registrar-agnostic version of TypeRegistrar.get_base_types
    """
    for kusto_type in _KustoType:
        if kusto_type.is_type_of(obj):
            # The object is already a member of Kusto types
            return {kusto_type}
    for type_registrar in (_plain_expression, _aggregation_expression, _typed_column):
        base_types = type_registrar.inverse(obj)
        if len(base_types) > 0:
            break
    assert len(base_types) > 0, f"get_base_types called for unsupported type: {type(obj).__name__}"
    return base_types



from threading import Thread, Lock
from typing import Any, Callable, List
from unittest.mock import patch

from pykusto import PyKustoClient, Query
# noinspection PyProtectedMember
from pykusto._src.client_base import Database
# noinspection PyProtectedMember
from pykusto._src.expressions import _StringColumn, _NumberColumn, _AnyTypeColumn, _BooleanColumn
# noinspection PyProtectedMember
from pykusto._src.type_utils import _KustoType
from test.test_base import TestBase, MockKustoClient, RecordedQuery, mock_tables_response, mock_getschema_response, mock_databases_response

background_query_lock = Lock()


class TestClientFetch(TestBase):
    query_thread: Thread = None
    query_results: List = []

    @staticmethod
    def query_in_background(query: Callable[[], Any]):
        with background_query_lock:
            assert TestClientFetch.query_thread is None
            TestClientFetch.query_results.clear()
            TestClientFetch.query_thread = Thread(target=lambda: TestClientFetch.query_results.extend(query()))
            TestClientFetch.query_thread.start()

    @staticmethod
    def get_background_query_result():
        with background_query_lock:
            assert TestClientFetch.query_thread is not None
            TestClientFetch.query_thread.join()
            TestClientFetch.query_thread = None
            return tuple(TestClientFetch.query_results)

    def test_column_fetch(self):
        mock_client = MockKustoClient(record_metadata=True)
        table = PyKustoClient(mock_client, fetch_by_default=False)['test_db']['mock_table']
        table.blocking_refresh()
        # Fetch query
        self.assertEqual(
            [RecordedQuery('test_db', '.show table mock_table | project AttributeName, AttributeType | limit 10000')],
            mock_client.recorded_queries,
        )
        # Dot notation
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        # Bracket notation
        self.assertType(table['foo'], _StringColumn)
        self.assertType(table['bar'], _NumberColumn)
        self.assertType(table['baz'], _AnyTypeColumn)

    def test_block_until_fetch_is_done(self):
        mock_client = MockKustoClient(block=True, record_metadata=True)
        client = PyKustoClient(mock_client)
        self.query_in_background(client.get_databases_names)
        mock_client.release()
        client.wait_for_items()
        # Make sure the fetch query was indeed called
        assert not mock_client.blocked()
        self.assertEqual(self.get_background_query_result(), ('test_db', ))

    def test_dir_before_fetch_is_done(self):
        mock_client = MockKustoClient(block=True, record_metadata=True)
        client = PyKustoClient(mock_client)
        self.query_in_background(lambda: dir(client))
        # Return the fetch
        mock_client.release()
        client.wait_for_items()
        # Make sure the fetch query was indeed called
        assert not mock_client.blocked()
        self.assertIn('test_db', self.get_background_query_result())

    def test_column_fetch_slow(self):
        mock_client = MockKustoClient(block=True)
        table = PyKustoClient(mock_client, fetch_by_default=False)['test_db']['mock_table']
        table.refresh()
        mock_client.wait_until_blocked()
        self.assertType(table['foo'], _AnyTypeColumn)
        self.assertType(table['bar'], _AnyTypeColumn)
        self.assertType(table['baz'], _AnyTypeColumn)
        # Return the fetch
        mock_client.release()
        table.wait_for_items()
        # Make sure the fetch query was indeed called
        assert not mock_client.blocked()

    @patch("pykusto._src.item_fetcher._DEFAULT_GET_ITEM_TIMEOUT_SECONDS", 0.0001)
    def test_table_fetch_slower_than_timeout(self):
        mock_client = MockKustoClient(block=True)
        try:
            PyKustoClient(mock_client)['test_db']['mock_table']
        finally:
            # # Return the fetch
            mock_client.release()

    def test_query_before_fetch_returned(self):
        mock_client = MockKustoClient(block=True, record_metadata=True)
        table = PyKustoClient(mock_client, fetch_by_default=False)['test_db']['mock_table']
        table.refresh()
        mock_client.wait_until_blocked()
        mock_client.do_not_block_next_requests()
        self.query_in_background(Query(table).take(5).execute)
        # Return the fetch
        mock_client.release()
        table.wait_for_items()
        self.get_background_query_result()
        # Make sure the fetch query was indeed called
        assert not mock_client.blocked()
        # Before the fix the order of returned query was reversed
        self.assertEqual(
            [
                RecordedQuery('test_db', '.show table mock_table | project AttributeName, AttributeType | limit 10000'),
                RecordedQuery('test_db', 'mock_table | take 5'),
            ],
            mock_client.recorded_queries,
        )

    def test_table_fetch(self):
        mock_client = MockKustoClient(record_metadata=True)
        db = PyKustoClient(mock_client, fetch_by_default=False)['test_db']
        db.blocking_refresh()
        self.assertEqual(
            [RecordedQuery('test_db', '.show database schema | project TableName, ColumnName, ColumnType | limit 10000')],
            mock_client.recorded_queries,
        )
        table = db.mock_table
        # Table columns
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        self.assertType(table['baz'], _AnyTypeColumn)
        # Bracket notation
        self.assertType(db['other_table']['foo'], _AnyTypeColumn)
        # Dot notation error
        self.assertRaises(
            AttributeError("PyKustoClient('test_cluster.kusto.windows.net').Database('test_db') has no attribute 'test_table_1'"),
            lambda: db.test_table_1
        )

    def test_two_tables_fetch(self):
        mock_client = MockKustoClient(record_metadata=True)
        db = PyKustoClient(mock_client, fetch_by_default=False)['test_db']
        db.blocking_refresh()
        self.assertEqual(
            [RecordedQuery('test_db', '.show database schema | project TableName, ColumnName, ColumnType | limit 10000')],
            mock_client.recorded_queries,
        )
        # Table columns
        self.assertType(db.mock_table.foo, _StringColumn)
        self.assertType(db.mock_table.bar, _NumberColumn)
        self.assertType(db.mock_table_2['baz'], _BooleanColumn)
        self.assertType(db['other_table']['foo'], _AnyTypeColumn)
        # Union
        table = db.get_table('mock_table', 'mock_table_2')
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        self.assertType(table.baz, _BooleanColumn)
        # Wildcard
        table = db.get_table('mock_*')
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        self.assertType(table.baz, _BooleanColumn)

    def test_union_column_name_conflict(self):
        mock_client = MockKustoClient(
            tables_response=mock_tables_response([
                ('test_table_1', [('foo', _KustoType.STRING), ('bar', _KustoType.INT)]),
                ('test_table_2', [('foo', _KustoType.BOOL)])
            ]),
            getschema_response=mock_getschema_response([
                ('foo_string', _KustoType.STRING), ('bar', _KustoType.INT), ('foo_bool', _KustoType.BOOL)
            ]),
            record_metadata=True,
        )
        db = PyKustoClient(mock_client, fetch_by_default=False)['test_db']
        db.blocking_refresh()
        table = db.get_table('test_table_*')
        table.blocking_refresh()  # To trigger name conflict resolution
        self.assertEqual(
            [
                # First trying the usual fetch
                RecordedQuery('test_db', '.show database schema | project TableName, ColumnName, ColumnType | limit 10000'),
                # Fallback for name conflict resolution
                RecordedQuery('test_db', 'union test_table_* | getschema | project ColumnName, DataType | limit 10000')
            ],
            mock_client.recorded_queries,
        )
        self.assertType(table.foo_string, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        self.assertType(table.foo_bool, _BooleanColumn)

    def test_union_wildcard_one_table(self):
        mock_client = MockKustoClient(record_metadata=True)
        db = PyKustoClient(mock_client, fetch_by_default=False)['test_db']
        db.blocking_refresh()
        self.assertEqual(
            [RecordedQuery('test_db', '.show database schema | project TableName, ColumnName, ColumnType | limit 10000')],
            mock_client.recorded_queries,
        )
        table = db.get_table('mock_table_*')
        self.assertType(table.foo, _AnyTypeColumn)
        self.assertType(table.bar, _AnyTypeColumn)
        self.assertType(table['baz'], _BooleanColumn)

    def test_database_fetch(self):
        mock_client = MockKustoClient(record_metadata=True)
        client = PyKustoClient(mock_client)
        client.wait_for_items()
        self.assertEqual(
            [RecordedQuery('', '.show databases schema | project DatabaseName, TableName, ColumnName, ColumnType | limit 100000')],
            mock_client.recorded_queries,
        )
        # Table columns
        table = client.test_db.mock_table
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _NumberColumn)
        self.assertType(table['baz'], _AnyTypeColumn)
        self.assertType(client.test_db['other_table']['foo'], _AnyTypeColumn)
        # Various utility methods
        db = client.get_database('test_db')
        self.assertType(db, Database)
        self.assertEqual('test_db', db.get_name())
        self.assertEqual(('test_db',), tuple(client.get_databases_names()))
        self.assertEqual(('mock_table', 'other_table'), tuple(client.test_db.get_table_names()))
        self.assertEqual(('foo', 'bar', 'baz'), tuple(client.test_db.mock_table.get_columns_names()))
        self.assertTrue({'foo', 'bar'} < set(dir(client.test_db.mock_table)))
        self.assertEqual("PyKustoClient('test_cluster.kusto.windows.net').Database('test_db').Table('mock_table')", repr(client.test_db.mock_table))

    def test_autocomplete_with_dot(self):
        mock_client = MockKustoClient(
            databases_response=mock_databases_response([('test_db', [('mock_table', [('foo', _KustoType.STRING), ('bar.baz', _KustoType.INT)])])]),
        )
        client = PyKustoClient(mock_client)
        client.wait_for_items()
        # Table columns
        table = client.test_db.mock_table
        self.assertType(table.foo, _StringColumn)
        self.assertType(table.bar, _AnyTypeColumn)
        self.assertType(table['bar.baz'], _NumberColumn)
        autocomplete_list = set(dir(client.test_db.mock_table))
        self.assertIn('foo', autocomplete_list)
        self.assertNotIn('bar.baz', autocomplete_list)

    def test_exception_from_autocomplete(self):
        mock_client = MockKustoClient(databases_response=self.raise_mock_exception)
        client = PyKustoClient(mock_client, fetch_by_default=True)
        autocomplete_list = set(dir(client))
        self.assertNotIn('test_db', autocomplete_list)

    def test_empty_database(self):
        mock_client = MockKustoClient(
            databases_response=mock_databases_response([
                ('test_db', [('mock_table', [('foo', _KustoType.STRING), ('bar', _KustoType.INT)])]),
                ('', [('test_table1', [('foo1', _KustoType.STRING), ('bar1', _KustoType.INT)])])
            ]),
            record_metadata=True,
        )
        client = PyKustoClient(mock_client)
        client.wait_for_items()
        self.assertEqual(
            [RecordedQuery('', '.show databases schema | project DatabaseName, TableName, ColumnName, ColumnType | limit 100000')],
            mock_client.recorded_queries,
        )
        self.assertType(client.test_db.mock_table.foo, _StringColumn)

    def test_client_database_names_not_fetched(self):
        client = PyKustoClient(MockKustoClient(), fetch_by_default=False)
        self.assertEqual(frozenset(['test_db']), set(client.get_databases_names()))

    def test_client_databases_not_fetched(self):
        client = PyKustoClient(MockKustoClient(), fetch_by_default=False)
        self.assertEqual(frozenset(['test_db']), set(db.get_name() for db in client.get_databases()))

    def test_exception_while_fetching(self):
        client = PyKustoClient(MockKustoClient(databases_response=self.raise_mock_exception))
        self.assertRaises(
            Exception("Mock exception"),
            lambda: set(client.get_databases_names()),
        )



import logging
import re
from datetime import datetime, timedelta

from pykusto import column_generator as col, Functions as f, Query
# noinspection PyProtectedMember
from pykusto._src.logger import _logger
from test.test_base import TestBase, test_logger
from test.test_base import mock_table as t


class TestFunction(TestBase):
    def test_acos(self):
        self.assertEqual(
            " | where (acos(numField)) > 4",
            Query().where(f.acos(t.numField) > 4).render()
        )

    def test_ago(self):
        self.assertEqual(
            " | where dateField > (ago(time(4.0:0:0.0)))",
            Query().where(t.dateField > f.ago(timedelta(4))).render()
        )
        self.assertEqual(
            " | where dateField > (ago(timespanField))",
            Query().where(t.dateField > f.ago(t.timespanField)).render()
        )

    def test_array_length(self):
        self.assertEqual(
            " | where (array_length(arrayField)) > 4",
            Query().where(f.array_length(t.arrayField) > 4).render()
        )

    def test_bag_keys(self):
        self.assertEqual(
            " | where (array_length(bag_keys(mapField))) > 4",
            Query().where(f.bag_keys(t.mapField).array_length() > 4).render()
        )

    def test_case(self):
        self.assertEqual(
            ' | extend bucket = case(numField <= 3, "Small", numField <= 10, "Medium", "Large")',
            Query().extend(bucket=f.case(t.numField <= 3, "Small", t.numField <= 10, 'Medium', 'Large')).render()
        )

    def test_ceiling(self):
        self.assertEqual(
            " | where (ceiling(numField)) > 4",
            Query().where(f.ceiling(t.numField) > 4).render()
        )

    def test_cos(self):
        self.assertEqual(
            " | where (cos(numField)) > 4",
            Query().where(f.cos(t.numField) > 4).render()
        )

    def test_endofday(self):
        self.assertEqual(
            " | where (endofday(dateField)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_day(t.dateField) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofday(dateField, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_day(t.dateField, 2) > datetime(2019, 7, 23)).render()
        )

    def test_endofmonth(self):
        self.assertEqual(
            " | where (endofmonth(dateField)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_month(t.dateField) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofmonth(dateField, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_month(t.dateField, 2) > datetime(2019, 7, 23)).render()
        )

    def test_endofweek(self):
        self.assertEqual(
            " | where (endofweek(dateField)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_week(t.dateField) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofweek(dateField, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_week(t.dateField, 2) > datetime(2019, 7, 23)).render()
        )

    def test_endofyear(self):
        self.assertEqual(
            " | where (endofyear(dateField)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_year(t.dateField) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofyear(dateField, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.end_of_year(t.dateField, 2) > datetime(2019, 7, 23)).render()
        )

    def test_dayofweek(self):
        self.assertEqual(
            " | where (dayofweek(datetime(2019-07-23 00:00:00.000000))) == time(1.0:0:0.0)",
            Query().where(f.day_of_week(datetime(2019, 7, 23)) == timedelta(1)).render()
        )

        self.assertEqual(
            " | where (dayofweek(dateField)) == time(1.0:0:0.0)",
            Query().where(f.day_of_week(t.dateField) == timedelta(1)).render()
        )

    def test_exp(self):
        self.assertEqual(
            " | where (exp(numField)) > 4",
            Query().where(f.exp(t.numField) > 4).render()
        )

    def test_exp10(self):
        self.assertEqual(
            " | where (exp10(numField)) > 4",
            Query().where(f.exp10(t.numField) > 4).render()
        )

    def test_exp2(self):
        self.assertEqual(
            " | where (exp2(numField)) > 4",
            Query().where(f.exp2(t.numField) > 4).render()
        )

    def test_extract(self):
        self.assertEqual(
            " | extend extract(@\"Duration=([0-9.]+)\", 1, stringField)",
            Query().extend(f.extract(r"Duration=([0-9.]+)", 1, t.stringField)).render()
        )

    def test_extract_type_literal(self):
        self.assertEqual(
            r' | extend extract(@"Duration=([0-9.]+)", 1, stringField, typeof(real))',
            Query().extend(f.extract(r"Duration=([0-9.]+)", 1, t.stringField, float)).render()
        )

    def test_extract_compiled_pattern(self):
        pat = re.compile(r"Duration=([0-9.]+)")
        self.assertEqual(
            r' | extend extract(@"Duration=([0-9.]+)", 1, stringField)',
            Query().extend(f.extract(pat, 1, t.stringField)).render()
        )

    def test_extract_all(self):
        self.assertEqual(
            r' | extend extract_all(@"([\da-f]{2})", stringField)',
            Query().extend(f.extract_all(r"([\da-f]{2})", t.stringField)).render()
        )

    def test_extract_all_capture_group(self):
        self.assertEqual(
            r' | extend extract_all(@"([\da-f]{2})", dynamic([1, 3]), stringField)',
            Query().extend(f.extract_all(r"([\da-f]{2})", t.stringField, [1, 3])).render()
        )

    def test_extract_all_named_capture_group(self):
        self.assertEqual(
            r' | extend extract_all(@"(?P<first>[\da-f]{2})", dynamic(["first", 3]), stringField)',
            Query().extend(f.extract_all(r"(?P<first>[\da-f]{2})", t.stringField, ["first", 3])).render()
        )

    def test_extract_all_compiled_pattern(self):
        pat = re.compile(r"([\da-f]{2})")
        self.assertEqual(
            r' | extend extract_all(@"([\da-f]{2})", stringField)',
            Query().extend(f.extract_all(pat, t.stringField)).render()
        )

    def test_floor(self):
        self.assertEqual(
            " | where (floor(dateField, time(0.12:0:0.0))) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.floor(t.dateField, timedelta(0.5)) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (floor(numField, 0.1)) > 3",
            Query().where(f.floor(t.numField, 0.1) > 3).render()
        )
        self.assertEqual(
            " | where (floor(numField, 0.1)) > 3",
            Query().where(t.numField.floor(0.1) > 3).render()
        )

    def test_format_datetime(self):
        self.assertEqual(
            ' | where (format_datetime(dateField, "yy-MM-dd [HH:mm:ss]")) == "2019-07-23 00:00:00"',
            Query().where(f.format_datetime(t.dateField, 'yy-MM-dd [HH:mm:ss]') == '2019-07-23 00:00:00').render()
        )

    # noinspection SpellCheckingInspection
    def test_format_timespan(self):
        self.assertEqual(
            ' | where (format_timespan(timespanField, "h:m:s.fffffff")) == "2:3:4.1234500"',
            Query().where(f.format_timespan(t.timespanField, 'h:m:s.fffffff') == "2:3:4.1234500").render()
        )

    def test_getmonth(self):
        self.assertEqual(
            " | where (getmonth(dateField)) > 3",
            Query().where(f.get_month(t.dateField) > 3).render()
        )

    def test_gettype(self):
        self.assertEqual(
            ' | where (gettype(dateField)) == "datetime"',
            Query().where(f.get_type(t.dateField) == 'datetime').render()
        )

    def test_getyear(self):
        self.assertEqual(
            " | where (getyear(dateField)) > 2019",
            Query().where(f.get_year(t.dateField) > 2019).render()
        )

    def test_hourofday(self):
        self.assertEqual(
            " | where (hourofday(dateField)) == 3",
            Query().where(f.hour_of_day(t.dateField) == 3).render()
        )

    def test_hash(self):
        self.assertEqual(
            " | where (hash(stringField)) == 3",
            Query().where(f.hash(t.stringField) == 3).render()
        )
        self.assertEqual(
            " | where (hash(stringField, 100)) == 3",
            Query().where(f.hash(t.stringField, 100) == 3).render()
        )
        self.assertEqual(
            " | where (hash(stringField, numField)) == 3",
            Query().where(f.hash(t.stringField, t.numField) == 3).render()
        )

    def test_hash_sha256(self):
        self.assertEqual(
            " | where (hash_sha256(stringField)) == 3",
            Query().where(f.hash_sha256(t.stringField) == 3).render()
        )

    def test_isempty(self):
        self.assertEqual(
            " | where isempty(stringField)",
            Query().where(f.is_empty(t.stringField)).render()
        )

    def test_isfinite(self):
        self.assertEqual(
            " | where isfinite(numField)",
            Query().where(f.is_finite(t.numField)).render()
        )

    def test_isinf(self):
        self.assertEqual(
            " | where isinf(numField)",
            Query().where(f.is_inf(t.numField)).render()
        )

    def test_isnan(self):
        self.assertEqual(
            " | where isnan(numField)",
            Query().where(f.is_nan(t.numField)).render()
        )

    def test_isnotempty(self):
        self.assertEqual(
            " | where isnotempty(stringField)",
            Query().where(f.is_not_empty(t.stringField)).render()
        )

    def test_isnotnull(self):
        self.assertEqual(
            " | where isnotnull(stringField)",
            Query().where(f.is_not_null(t.stringField)).render()
        )

    def test_isnull(self):
        self.assertEqual(
            " | where isnull(stringField)",
            Query().where(f.is_null(t.stringField)).render()
        )

    def test_isutf8(self):
        self.assertEqual(
            " | where isutf8(stringField)",
            Query().where(f.is_utf8(t.stringField)).render()
        )

    def test_log(self):
        self.assertEqual(
            " | where (log(numField)) < 3",
            Query().where(f.log(t.numField) < 3).render()
        )

    def test_log10(self):
        self.assertEqual(
            " | where (log10(numField)) < 3",
            Query().where(f.log10(t.numField) < 3).render(),
        )

    def test_log2(self):
        self.assertEqual(
            " | where (log2(numField)) < 3",
            Query().where(f.log2(t.numField) < 3).render()
        )

    def test_loggamma(self):
        self.assertEqual(
            " | where (loggamma(numField)) < 3",
            Query().where(f.log_gamma(t.numField) < 3).render()
        )

    def test_make_datetime(self):
        self.assertEqual(
            " | where (make_datetime(numField, numField2, numField3, 0, 0, 0)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.make_datetime(t.numField, t.numField2, t.numField3) > datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (make_datetime(numField, numField2, numField3, numField4, 0, 0)) > datetime(2019-07-23 05:00:00.000000)",
            Query().where(f.make_datetime(t.numField, t.numField2, t.numField3, t.numField4) > datetime(2019, 7, 23, 5)).render()
        )
        self.assertEqual(
            " | where (make_datetime(numField, numField2, numField3, numField4, numField5, 0)) > datetime(2019-07-23 05:29:00.000000)",
            Query().where(
                f.make_datetime(t.numField, t.numField2, t.numField3, t.numField4, t.numField5) > datetime(2019, 7, 23, 5, 29)).render()
        )
        self.assertEqual(
            " | where (make_datetime(numField, numField2, numField3, numField4, numField5, numField6)) > datetime(2019-07-23 05:29:15.000000)",
            Query().where(
                f.make_datetime(t.numField, t.numField2, t.numField3, t.numField4, t.numField5, t.numField6) > datetime(2019, 7, 23, 5, 29, 15)).render()
        )

    def test_make_string(self):
        self.assertEqual(
            " | where (make_string(75, 117, 115, 116, 111)) == \"Kusto\"",
            Query().where(f.make_string(75, 117, 115, 116, 111) == 'Kusto').render()
        )
        self.assertEqual(
            " | where (make_string(dynamic([75, 117, 115]), 116, 111)) == \"Kusto\"",
            Query().where(f.make_string([75, 117, 115], 116, 111) == 'Kusto').render()
        )

    def test_month_of_year(self):
        self.assertEqual(
            " | where (monthofyear(dateField)) == 12",
            Query().where(f.month_of_year(t.dateField) == 12).render()
        )
        self.assertEqual(
            " | where (monthofyear(datetime(2015-12-14 00:00:00.000000))) == 12",
            Query().where(f.month_of_year(datetime(2015, 12, 14)) == 12).render()
        )

    def test_now(self):
        self.assertEqual(
            " | where dateField < (now())",
            Query().where(t.dateField < f.now()).render()
        )
        self.assertEqual(
            " | where dateField < (now(time(-3.0:0:0.0)))",
            Query().where(t.dateField < f.now(timedelta(-3))).render()
        )

    def test_pack_all(self):
        self.assertEqual(
            " | extend packed = pack_all()",
            Query().extend(packed=f.pack_all()).render()
        )

    def test_pack_dictionary(self):
        self.assertEqual(
            ' | extend foo = pack_dictionary("bar", numField, "baz", stringField)',
            Query().extend(foo=f.pack_dictionary(bar=t.numField, baz=t.stringField)).render()
        )

    def test_parse_json_to_string(self):
        self.assertEqual(
            ' | where (tostring(parse_json(stringField))) contains "ABC"',
            Query().where(f.parse_json(t.stringField).to_string().contains('ABC')).render()
        )

    def test_parse_json_brackets(self):
        self.assertEqual(
            ' | where (tostring(parse_json(stringField)["bar"])) contains "ABC"',
            Query().where(f.parse_json(t.stringField)['bar'].to_string().contains('ABC')).render()
        )

    def test_parse_json_dot(self):
        self.assertEqual(
            ' | where (tostring(parse_json(stringField).bar)) contains "ABC"',
            Query().where(f.parse_json(t.stringField).bar.to_string().contains('ABC')).render()
        )

    def test_parse_json_number_expression(self):
        self.assertEqual(
            ' | where (todouble(parse_json(stringField).bar)) > 4',
            Query().where(f.to_double(f.parse_json(t.stringField).bar) > 4).render()
        )

    def test_parse_json_array(self):
        self.assertEqual(
            ' | where (parse_json(stringField)[2]) == 3',
            Query().where(f.parse_json(t.stringField)[2] == 3).render()
        )

    def test_parse_json_nesting(self):
        self.assertEqual(
            ' | where (parse_json(stringField)["a"].b[2]) contains "bar"',
            Query().where(f.parse_json(t.stringField)['a'].b[2].contains('bar')).render())

    def test_pow(self):
        self.assertEqual(
            " | where (pow(numField, numField2)) > 3",
            Query().where(f.pow(t.numField, t.numField2) > 3).render()
        )

    def test_round(self):
        self.assertEqual(
            " | where (round(numField, numField2)) == 3",
            Query().where(f.round(t.numField, t.numField2) == 3).render()
        )

    def test_sign(self):
        self.assertEqual(
            " | where (sign(numField)) == 1",
            Query().where(f.sign(t.numField) == 1).render()
        )

    def test_sqrt(self):
        self.assertEqual(
            " | where (sqrt(numField)) > 1",
            Query().where(f.sqrt(t.numField) > 1).render()
        )

    def test_startofday(self):
        self.assertEqual(
            " | where (startofday(dateField)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.start_of_day(t.dateField) > datetime(2019, 7, 23)).render()
        )

    def test_startofmonth(self):
        self.assertEqual(
            " | where (startofmonth(dateField)) > datetime(2019-07-01 00:00:00.000000)",
            Query().where(f.start_of_month(t.dateField) > datetime(2019, 7, 1)).render()
        )

    def test_startofweek(self):
        self.assertEqual(
            " | where (startofweek(dateField)) > datetime(2019-07-08 00:00:00.000000)",
            Query().where(f.start_of_week(t.dateField) > datetime(2019, 7, 8)).render()
        )

    def test_startofyear(self):
        self.assertEqual(
            " | where (startofyear(dateField)) > datetime(2019-01-01 00:00:00.000000)",
            Query().where(f.start_of_year(t.dateField) > datetime(2019, 1, 1)).render()
        )

    def test_strcat(self):
        self.assertEqual(
            ' | extend strcat("hello", ",", stringField, "!")',
            Query().extend(f.strcat("hello", ',', t.stringField, '!')).render()
        )
        self.assertEqual(
            ' | extend strcat(stringField, "!")',
            Query().extend(f.strcat(t.stringField, '!')).render()
        )

    def test_strcat_one_argument(self):
        self.assertRaises(
            ValueError("strcat requires at least two arguments"),
            lambda: f.strcat(t.stringField)
        )

    def test_strcat_delim(self):
        self.assertEqual(
            ' | extend strcat_delim("-", "hello", ",", stringField, "!")',
            Query().extend(f.strcat_delim('-', "hello", ',', t.stringField, '!')).render()
        )
        self.assertEqual(
            ' | extend strcat_delim("-", ",", stringField)',
            Query().extend(f.strcat_delim('-', ',', t.stringField)).render()
        )

    def test_strcat_array(self):
        self.assertEqual(
            ' | where (strcat_array(arrayField, ",")) == "A,B,C"',
            Query().where(f.strcat_array(t.arrayField, ',') == 'A,B,C').render()
        )
        self.assertEqual(
            ' | where (strcat_array(dynamic(["A", "B", "C"]), ",")) == "A,B,C"',
            Query().where(f.strcat_array(['A', 'B', 'C'], ',') == 'A,B,C').render()
        )

    def test_strcmp(self):
        self.assertEqual(
            " | where (strcmp(stringField, stringField2)) == 1",
            Query().where(f.strcmp(t.stringField, t.stringField2) == 1).render()
        )

    def test_string_size(self):
        self.assertEqual(
            " | where (string_size(stringField)) == 1",
            Query().where(f.string_size(t.stringField) == 1).render()
        )

    def test_strlen(self):
        self.assertEqual(
            " | where (strlen(stringField)) == 1",
            Query().where(f.strlen(t.stringField) == 1).render()
        )

    def test_strrep(self):
        self.assertEqual(
            ' | where (strrep(stringField, numField)) == "ABC"',
            Query().where(f.strrep(t.stringField, t.numField) == 'ABC').render()
        )
        self.assertEqual(
            ' | where (strrep(stringField, numField, ",")) == "ABC,ABC"',
            Query().where(f.strrep(t.stringField, t.numField, ',') == 'ABC,ABC').render()
        )
        self.assertEqual(
            ' | where (strrep(stringField, numField, stringField2)) == "ABC,ABC"',
            Query().where(f.strrep(t.stringField, t.numField, t.stringField2) == 'ABC,ABC').render()
        )

    def test_substring(self):
        self.assertEqual(
            ' | where (substring(stringField, numField)) == "ABC"',
            Query().where(f.substring(t.stringField, t.numField) == 'ABC').render()
        )
        self.assertEqual(
            ' | where (substring(stringField, numField, 4)) == "ABC,ABC"',
            Query().where(f.substring(t.stringField, t.numField, 4) == 'ABC,ABC').render()
        )

    def test_split(self):
        self.assertEqual(
            ' | extend foo = split(stringField, "_", 3)',
            Query().extend(foo=f.split(t.stringField, "_", 3)).render()
        )
        self.assertEqual(
            ' | extend foo = split(stringField, "_")[3]',
            Query().extend(foo=f.split(t.stringField, "_")[3]).render()
        )
        self.assertEqual(
            ' | extend foo = split("1_2", "_")[3]',
            Query().extend(foo=f.split("1_2", "_")[3]).render()
        )

    def test_count_of(self):
        self.assertEqual(
            ' | where (countof(stringField, "abc", "normal")) == 2',
            Query().where(f.count_of(t.stringField, "abc") == 2).render()
        )

    def test_tobool(self):
        self.assertEqual(
            " | where tobool(stringField)",
            Query().where(f.to_bool(t.stringField)).render()
        )

    def test_todatetime(self):
        self.assertEqual(
            " | extend foo = todatetime(stringField)",
            Query().extend(foo=f.to_datetime(t.stringField)).render()
        )

    def test_todecimal(self):
        self.assertEqual(
            " | where (todecimal(stringField)) > 0.2",
            Query().where(f.to_decimal(t.stringField) > 0.2).render()
        )

    def test_todouble(self):
        self.assertEqual(
            " | where (todouble(stringField)) > 0.2",
            Query().where(f.to_double(t.stringField) > 0.2).render()
        )

    def test_to_dynamic(self):
        self.assertEqual(
            " | extend dyn = todynamic(stringField)",
            Query().extend(dyn=f.to_dynamic(t.stringField)).render()
        )

    def test_toint(self):
        self.assertEqual(
            " | where (toint(stringField)) > 1",
            Query().where(f.to_int(t.stringField) > 1).render()
        )

    def test_tolong(self):
        self.assertEqual(
            " | where (tolong(stringField)) > 2222222222",
            Query().where(f.to_long(t.stringField) > 2222222222).render()
        )

    def test_tolower(self):
        self.assertEqual(
            ' | where (tolower(stringField)) == "foo"',
            Query().where(f.to_lower(t.stringField) == "foo").render()
        )

    def test_toreal(self):
        self.assertEqual(
            " | where (toreal(stringField)) > 0.2",
            Query().where(f.to_real(t.stringField) > 0.2).render()
        )

    def test_to_timespan(self):
        self.assertEqual(
            " | extend t = totimespan(stringField)",
            Query().extend(t=f.to_timespan(t.stringField)).render()
        )

    def test_toupper(self):
        self.assertEqual(
            ' | where (toupper(stringField)) == "FOO"',
            Query().where(f.to_upper(t.stringField) == "FOO").render()
        )

    def test_tohex(self):
        self.assertEqual(
            ' | where (tohex(256)) == "100"',
            Query().where(f.to_hex(256) == "100").render()
        )

    def test_trim(self):
        self.assertEqual(
            ' | where (trim("--", stringField)) == "text"',
            Query().where(f.trim("--", t.stringField) == "text").render()
        )

    def test_rand(self):
        self.assertEqual(
            " | extend rnd = rand()",
            Query().extend(rnd=f.rand()).render()
        )
        self.assertEqual(
            " | extend rnd = rand(5)",
            Query().extend(rnd=f.rand(5)).render()
        )

    # ------------------------------------------------------
    # Aggregation Functions
    # ------------------------------------------------------

    def test_any(self):
        self.assertEqual(
            " | summarize any(stringField, numField, boolField)",
            Query().summarize(f.any(t.stringField, t.numField, t.boolField)).render()
        )

    def test_any_wildcard(self):
        self.assertEqual(
            " | summarize any(*)",
            Query().summarize(f.any()).render()
        )

    def test_any_if(self):
        self.assertEqual(
            " | summarize anyif(stringField, boolField)",
            Query().summarize(f.any_if(t.stringField, t.boolField)).render()
        )

    def test_aggregation_assign_to(self):
        self.assertEqual(
            " | summarize foo = any(stringField)",
            Query().summarize(f.any(t.stringField).assign_to(col.foo)).render()
        )

    def test_aggregation_by_assign_to(self):
        self.assertEqual(
            " | summarize any(stringField) by foo = boolField",
            Query().summarize(f.any(t.stringField)).by(t.boolField.assign_to(col.foo)).render()
        )

    def test_aggregation_assign_to_multiple(self):
        self.assertRaises(
            ValueError("Aggregations cannot be assigned to multiple columns"),
            lambda: f.any(t.stringField).assign_to(col.foo, col.bar)
        )

    def test_arg_max(self):
        self.assertEqual(
            " | summarize arg_max(stringField, numField, boolField)",
            Query().summarize(f.arg_max(t.stringField, t.numField, t.boolField)).render()
        )

    def test_arg_max_wildcard(self):
        self.assertEqual(
            " | summarize arg_max(stringField, *)",
            Query().summarize(f.arg_max(t.stringField)).render()
        )

    def test_arg_min(self):
        self.assertEqual(
            " | summarize arg_min(stringField, numField, boolField)",
            Query().summarize(f.arg_min(t.stringField, t.numField, t.boolField)).render()
        )

    def test_arg_min_wildcard(self):
        self.assertEqual(
            " | summarize arg_min(stringField, *)",
            Query().summarize(f.arg_min(t.stringField)).render()
        )

    def test_avg(self):
        self.assertEqual(
            " | summarize avg(numField)",
            Query().summarize(f.avg(t.numField)).render()
        )
        self.assertEqual(
            " | summarize avg(numField) - 5",
            Query().summarize(f.avg(t.numField) - 5).render()
        )

    def test_avgif(self):
        self.assertEqual(
            " | summarize avgif(numField, boolField)",
            Query().summarize(f.avg_if(t.numField, t.boolField)).render()
        )

    def test_bin(self):
        self.assertEqual(
            " | summarize avg(numField) by bin(numField2, 0.1)",
            Query().summarize(f.avg(t.numField)).by(f.bin(t.numField2, 0.1)).render()
        )
        self.assertEqual(
            " | summarize avg(numField2) by bin(dateField, time(0.12:0:0.0))",
            Query().summarize(f.avg(t.numField2)).by(f.bin(t.dateField, timedelta(0.5))).render()
        )
        self.assertEqual(
            " | summarize avg(numField2) by bin(timespanField, time(0.12:0:0.0))",
            Query().summarize(f.avg(t.numField2)).by(f.bin(t.timespanField, timedelta(0.5))).render()
        )

    def test_bin_at(self):
        self.assertEqual(
            " | summarize avg(numField) by bin_at(numField2, 0.1, 1)",
            Query().summarize(f.avg(t.numField)).by(f.bin_at(t.numField2, 0.1, 1)).render()
        )
        self.assertEqual(
            " | summarize avg(numField) by bin_at(dateField, time(0.12:0:0.0), datetime(2019-07-08 00:00:00.000000))",
            Query().summarize(f.avg(t.numField)).by(f.bin_at(t.dateField, timedelta(0.5), datetime(2019, 7, 8))).render()
        )
        self.assertEqual(
            " | summarize avg(numField) by bin_at(timespanField, time(0.12:0:0.0), time(1.0:0:0.0))",
            Query().summarize(f.avg(t.numField)).by(f.bin_at(t.timespanField, timedelta(0.5), timedelta(days=1))).render()
        )

    def test_bin_auto(self):
        self.assertEqual(
            " | summarize avg(numField) by bin_auto(numField)",
            Query().summarize(f.avg(t.numField)).by(f.bin_auto(t.numField)).render()
        )
        self.assertEqual(
            " | summarize avg(numField) by bin_auto(timespanField)",
            Query().summarize(f.avg(t.numField)).by(f.bin_auto(t.timespanField)).render()
        )

    def test_count(self):
        self.assertEqual(
            " | summarize count()",
            Query().summarize(f.count()).render()
        )
        self.assertEqual(
            " | summarize count(stringField)",
            Query().summarize(f.count(t.stringField)).render()
        )

    def test_countif(self):
        self.assertEqual(
            " | summarize countif(numField == 1)",
            Query().summarize(f.count_if(t.numField == 1)).render()
        )

    def test_dcount(self):
        self.assertEqual(
            " | summarize dcount(stringField)",
            Query().summarize(f.dcount(t.stringField)).render()
        )
        acc = 0.1
        self.assertEqual(
            " | summarize dcount(numField, 0.1)",
            Query().summarize(f.dcount(t.numField, acc)).render()
        )

    def test_dcountif(self):
        self.assertEqual(
            " | summarize dcountif(stringField, boolField, 1)",
            Query().summarize(f.dcount_if(t.stringField, t.boolField)).render()
        )

    def test_make_bag(self):
        self.assertEqual(
            " | summarize make_bag(stringField)",
            Query().summarize(f.make_bag(t.stringField)).render()
        )
        self.assertEqual(
            " | summarize make_bag(stringField, numField)",
            Query().summarize(f.make_bag(t.stringField, t.numField)).render()
        )

    def test_make_list(self):
        self.assertEqual(
            " | summarize make_list(stringField)",
            Query().summarize(f.make_list(t.stringField)).render()
        )
        self.assertEqual(
            " | summarize make_list(stringField, numField)",
            Query().summarize(f.make_list(t.stringField, t.numField)).render()
        )

    def test_make_set(self):
        self.assertEqual(
            " | summarize make_set(stringField)",
            Query().summarize(f.make_set(t.stringField)).render()
        )
        self.assertEqual(
            " | summarize make_set(stringField, numField)",
            Query().summarize(f.make_set(t.stringField, t.numField)).render()
        )

    def test_take_any_single_expr(self):
        self.assertEqual(
            " | summarize take_any(numField)",
            Query().summarize(f.take_any(t.numField)).render()
        )

    def test_take_any_multiple_expr(self):
        self.assertEqual(
            " | summarize take_any(numField, stringField, boolField)",
            Query().summarize(f.take_any(t.numField, t.stringField, t.boolField)).render()
        )

    def test_take_any_all(self):
        self.assertEqual(
            " | summarize take_any(*)",
            Query().summarize(f.take_any_all()).render()
        )

    def test_max(self):
        self.assertEqual(
            " | summarize max(numField)",
            Query().summarize(f.max(t.numField)).render()
        )

    def test_min(self):
        self.assertEqual(
            " | summarize min(numField)",
            Query().summarize(f.min(t.numField)).render()
        )

    def test_max_if(self):
        self.assertEqual(
            " | summarize maxif(numField, boolField)",
            Query().summarize(f.max_if(t.numField, t.boolField)).render()
        )

    def test_min_if(self):
        self.assertEqual(
            " | summarize minif(numField, boolField)",
            Query().summarize(f.min_if(t.numField, t.boolField)).render()
        )

    def test_percentile(self):
        self.assertEqual(
            " | summarize percentiles(numField, 5)",
            Query().summarize(f.percentile(t.numField, 5)).render()
        )

    def test_percentiles(self):
        self.assertEqual(
            " | summarize percentiles(numField, 5, 50, 95)",
            Query().summarize(f.percentiles(t.numField, 5, 50, 95)).render()
        )

    def test_percentiles_array(self):
        self.assertEqual(
            " | summarize percentiles_array(numField, 5, 50, 95)",
            Query().summarize(f.percentiles_array(t.numField, 5, 50, 95)).render()
        )

    def test_stdev(self):
        self.assertEqual(
            " | summarize stdev(numField)",
            Query().summarize(f.stdev(t.numField)).render()
        )

    def test_stdevif(self):
        self.assertEqual(
            " | summarize stdevif(numField, boolField)",
            Query().summarize(f.stdevif(t.numField, t.boolField)).render()
        )

    def test_stdevp(self):
        self.assertEqual(
            " | summarize stdevp(numField)",
            Query().summarize(f.stdevp(t.numField)).render()
        )

    def test_sum(self):
        self.assertEqual(
            " | summarize sum(numField)",
            Query().summarize(f.sum(t.numField)).render()
        )

    def test_sumif(self):
        self.assertEqual(
            " | summarize sumif(numField, boolField)",
            Query().summarize(f.sum_if(t.numField, t.boolField)).render()
        )

    def test_variance(self):
        self.assertEqual(
            " | summarize variance(numField)",
            Query().summarize(f.variance(t.numField)).render()
        )

    def test_varianceif(self):
        self.assertEqual(
            " | summarize varianceif(numField, boolField)",
            Query().summarize(f.variance_if(t.numField, t.boolField)).render()
        )

    def test_variancep(self):
        self.assertEqual(
            " | summarize variancep(numField)",
            Query().summarize(f.variancep(t.numField)).render()
        )

    def test_nesting(self):
        self.assertEqual(
            " | summarize active_days = dcount(bin(dateField, time(1.0:0:0.0)))",
            Query().summarize(active_days=f.dcount(f.bin(t.dateField, timedelta(1)))).render()
        )

    def test_iff(self):
        self.assertEqual(
            " | project foo = iff(dateField > (ago(time(2.0:0:0.0))), time(3.0:0:0.0), time(4.0:0:0.0))",
            Query().project(foo=f.iff(t.dateField > f.ago(timedelta(2)), timedelta(3), timedelta(4))).render()
        )

    def test_iff_expression_return_type(self):
        self.assertEqual(
            " | project foo = iff(dateField > (ago(time(2.0:0:0.0))), array_length(arrayField), array_length(arrayField2))",
            Query().project(foo=f.iff(t.dateField > f.ago(timedelta(2)), f.array_length(t.arrayField), f.array_length(t.arrayField2))).render()
        )

    def test_iff_different_types(self):
        with self.assertLogs(_logger, logging.WARN) as cm:
            self.assertEqual(
                ' | project foo = iff(dateField > (ago(time(2.0:0:0.0))), time(3.0:0:0.0), "hello")',
                Query().project(foo=f.iff(t.dateField > f.ago(timedelta(2)), timedelta(3), "hello")).render()
            )
        self.assertEqual(
            ['WARNING:pykusto:The second and third arguments must be of the same type, but they are: timespan and string. '
             'If this is a mistake, please report it at https://github.com/Azure/pykusto/issues'],
            cm.output
        )

    def test_iff_related_types(self):
        self.assertEqual(
            " | project foo = iff(dateField > (ago(time(2.0:0:0.0))), 2, array_length(arrayField))",
            Query().project(foo=f.iff(t.dateField > f.ago(timedelta(2)), 2, f.array_length(t.arrayField))).render()
        )

    def test_iff_ambiguous_type(self):
        with self.assertLogs(test_logger, logging.WARN) as cm:
            self.assertEqual(
                " | project foo = iff(boolField, time(3.0:0:0.0), foo - bar)",
                Query().project(foo=f.iff(t.boolField, timedelta(3), col.foo - col.bar)).render()
            )
        self.assertEqual([], cm.output)

    def test_iif(self):
        # iif is just an alias to iff
        self.assertEqual(
            " | project foo = iff(dateField > (ago(time(2.0:0:0.0))), time(3.0:0:0.0), time(4.0:0:0.0))",
            Query().project(foo=f.iif(t.dateField > f.ago(timedelta(2)), timedelta(3), timedelta(4))).render()
        )

    def test_pack(self):
        self.assertEqual(
            ' | extend foo = pack("bar", numField, "baz", stringField)',
            Query().extend(foo=f.pack(bar=t.numField, baz=t.stringField)).render()
        )

    def test_pack_array(self):
        self.assertEqual(
            ' | extend foo = pack_array(numField, stringField)',
            Query().extend(foo=f.pack_array(t.numField, t.stringField)).render()
        )

    def test_set_has_element(self):
        self.assertEqual(
            ' | extend foo = set_has_element(arrayField, numField)',
            Query().extend(foo=f.set_has_element(t.arrayField, t.numField)).render()
        )

    def test_set_difference(self):
        self.assertEqual(
            ' | extend foo = set_difference(arrayField, arrayField2)',
            Query().extend(foo=f.set_difference(t.arrayField, t.arrayField2)).render()
        )

    def test_set_intersect(self):
        self.assertEqual(
            ' | extend foo = set_intersect(arrayField, arrayField2)',
            Query().extend(foo=f.set_intersect(t.arrayField, t.arrayField2)).render()
        )

    def test_set_union(self):
        self.assertEqual(
            ' | extend foo = set_union(arrayField, arrayField2)',
            Query().extend(foo=f.set_union(t.arrayField, t.arrayField2)).render()
        )

    def test_array_concat(self):
        self.assertEqual(
            ' | extend foo = array_concat(arrayField, arrayField2)',
            Query().extend(foo=f.array_concat(t.arrayField, t.arrayField2)).render()
        )

    def test_array_iif(self):
        self.assertEqual(
            ' | extend foo = array_iif(arrayField, arrayField2, arrayField3)',
            Query().extend(foo=f.array_iif(t.arrayField, t.arrayField2, t.arrayField3)).render()
        )

    def test_array_index_of(self):
        self.assertEqual(
            ' | extend foo = array_index_of(arrayField, numField)',
            Query().extend(foo=f.array_index_of(t.arrayField, t.numField)).render()
        )

    def test_array_rotate_left(self):
        self.assertEqual(
            ' | extend foo = array_rotate_left(arrayField, numField)',
            Query().extend(foo=f.array_rotate_left(t.arrayField, t.numField)).render()
        )

    def test_array_rotate_right(self):
        self.assertEqual(
            ' | extend foo = array_rotate_right(arrayField, numField)',
            Query().extend(foo=f.array_rotate_right(t.arrayField, t.numField)).render()
        )

    def test_array_shift_left(self):
        self.assertEqual(
            ' | extend foo = array_shift_left(arrayField, numField)',
            Query().extend(foo=f.array_shift_left(t.arrayField, t.numField)).render()
        )

    def test_array_shift_right(self):
        self.assertEqual(
            ' | extend foo = array_shift_right(arrayField, numField, numField2)',
            Query().extend(foo=f.array_shift_right(t.arrayField, t.numField, t.numField2)).render()
        )

    def test_array_slice(self):
        self.assertEqual(
            ' | extend foo = array_slice(arrayField, numField, numField2)',
            Query().extend(foo=f.array_slice(t.arrayField, t.numField, t.numField2)).render()
        )

    def test_array_split(self):
        self.assertEqual(
            ' | extend foo = array_split(arrayField, numField)',
            Query().extend(foo=f.array_split(t.arrayField, t.numField)).render()
        )

    def test_ingestion_time(self):
        self.assertEqual(
            ' | extend ingestionTime = ingestion_time()',
            Query().extend(ingestionTime=f.ingestion_time()).render()
        )

    def test_all_of(self):
        self.assertEqual(
            ' | where boolField and (numField > numField2) and (stringField contains "hello")',
            Query().where(f.all_of(t.boolField, t.numField > t.numField2, t.stringField.contains('hello'))).render()
        )

    def test_any_of(self):
        self.assertEqual(
            ' | where boolField or (numField > numField2) or (stringField contains "hello")',
            Query().where(f.any_of(t.boolField, t.numField > t.numField2, t.stringField.contains('hello'))).render()
        )
